import 'dart:convert';
import 'dart:io';

import 'package:args/command_runner.dart';
import 'package:mi_service/src/crypto.dart';
import 'package:mi_service/src/utils.dart';

class DecodeCommand extends Command<void> {
  @override
  final String name = 'decode';

  @override
  final String description = 'Decode RC4 encrypted data for IO service.';

  late final String type;
  late final String? payload;
  late final String? nonce;
  late final String? ssecurity;

  DecodeCommand() {
    argParser
      ..addOption(
        'type',
        abbr: 't',
        help: 'Data type.',
        defaultsTo: 'req',
        allowed: ['req', 'resp'],
        callback: (s) => type = s!,
      )
      ..addOption(
        'payload',
        abbr: 'p',
        help: 'Raw req/resp payload.',
        callback: (s) => payload = s,
      )
      ..addOption(
        'nonce',
        abbr: 'n',
        help: 'The nonce used to encrypt the payload.',
        callback: (s) => nonce = s,
      )
      ..addOption(
        'ssecurity',
        abbr: 's',
        help: 'The ssecurity used to encrypt the payload.',
        callback: (s) => ssecurity = s,
      );
  }

  @override
  Future<void> run() async {
    if (payload == null) {
      logger.e('Payload is required.');
      printUsage();
      return;
    }

    if (type == 'req') {
      final params = Uri.parse('?$payload').queryParameters;

      logger.d('Request params:\n${jsonEncoder.convert(params)}');

      final data = params['data']!;
      final nonce = this.nonce ?? params['_nonce']!;
      final ssecurity = this.ssecurity ?? params['ssecurity']!;

      final signedNonce = signNonce(
        nonce: base64Decode(nonce),
        ssecurity: ssecurity,
      );

      final decrypted = utf8.decode(
        MiRc4Codec(signedNonce).convert(
          base64Decode(data),
        ),
      );

      print(jsonEncoder.convert(jsonDecode(decrypted)));
    } else {
      final nonce = this.nonce;
      final ssecurity = this.ssecurity;

      if (nonce == null || ssecurity == null) {
        logger.e('Nonce and ssecurity is required.');
        printUsage();
        return;
      }

      final signedNonce = signNonce(
        nonce: base64Decode(Uri.decodeComponent(nonce)),
        ssecurity: Uri.decodeComponent(ssecurity),
      );

      var decrypted = MiRc4Codec(signedNonce).convert(base64Decode(payload!));

      // Decompress if gzipped.
      if (decrypted[0] == 0x1F && decrypted[1] == 0x8B) {
        decrypted = gzip.decode(decrypted);
      }

      print(jsonEncoder.convert(jsonDecode(utf8.decode(decrypted))));
    }
  }
}
