import 'dart:convert';
import 'dart:typed_data';

import 'package:convert/convert.dart';
import 'package:crypto/crypto.dart';

import 'utils.dart';

/// Generate a nonce.
List<int> genNonce() {
  final bytes = ByteData(12)
    // 64 bits random.
    ..setUint32(0, random.nextInt(0xFFFFFFFF))
    ..setUint32(4, random.nextInt(0xFFFFFFFF))
    // 32 bits timestamp.
    ..setUint32(8, DateTime.now().millisecondsSinceEpoch ~/ 60000);
  return bytes.buffer.asUint8List();
}

/// Sign nonce using ssecurity.
List<int> signNonce({
  required List<int> nonce,
  required String ssecurity,
}) {
  final output = AccumulatorSink<Digest>();
  sha256.startChunkedConversion(output)
    ..add(base64.decode(ssecurity))
    ..add(nonce)
    ..close();
  return output.events.single.bytes;
}

/// Generate signature for request.
String genSignature({
  required List<int> signedNonce,
  required String url,
  required Map<String, dynamic> params,
  String method = 'POST',
}) {
  final msg = [
    method.toUpperCase(),
    Uri.parse(url).path.replaceFirst('/app', ''),
    for (final entry in params.entries) '${entry.key}=${entry.value}',
    base64Encode(signedNonce),
  ].join('&');

  return base64Encode(sha1.convert(utf8.encode(msg)).bytes);
}

/// The RC4 cipher used to encrypt & decrypt data for IO service.
class MiRc4Codec {
  final _s = List.generate(256, (i) => i);
  var _j = 0;
  var _k = 0;

  MiRc4Codec(List<int> key) {
    // KSA
    {
      var x = 0;
      for (var i = 0; i < 256; i++) {
        x = (x + key[i % key.length] + _s[i]) % 256;
        var temp = _s[i];
        _s[i] = _s[x];
        _s[x] = temp;
      }
    }

    // 1024 fake rounds.
    for (var i = 0; i < 1024; i++) {
      _moveOn();
    }
  }

  void _moveOn() {
    _j = (_j + 1) % 256;
    _k = (_k + _s[_j]) % 256;
    var temp = _s[_j];
    _s[_j] = _s[_k];
    _s[_k] = temp;
  }

  List<int> convert(List<int> data) {
    var output = Uint8List(data.length);
    for (var i = 0; i < data.length; i++) {
      _moveOn();
      output[i] = (data[i] ^ _s[(_s[_j] + _s[_k]) % 256]) % 256;
    }
    return output;
  }
}
