import 'dart:convert';
import 'dart:io';

import 'package:dio/adapter.dart';
import 'package:dio/dio.dart';

import 'crypto.dart';
import 'model/credential.dart';
import 'model/io.dart';
import 'utils.dart';

class MiIOService {
  final _dio = Dio(BaseOptions(
    baseUrl: 'https://api.io.mi.com',
    headers: {
      HttpHeaders.acceptHeader: '*/*',
      HttpHeaders.acceptEncodingHeader: 'identity',
      HttpHeaders.acceptLanguageHeader: 'en-US',
      'x-xiaomi-protocal-flag-cli': 'PROTOCAL-HTTP2',
      'miot-accept-encoding': 'GZIP',
      'miot-encrypt-algorithm': 'ENCRYPT-RC4',
    },
    contentType: 'application/x-www-form-urlencoded',
  ));

  final MiServiceCredential credential;

  MiIOService(
    this.credential, {
    String? deviceId,
    String? passportDeviceId,
  }) {
    final httpClientAdapter = _dio.httpClientAdapter;
    if (httpClientAdapter is DefaultHttpClientAdapter) {
      // Remove default UA.
      httpClientAdapter.onHttpClientCreate =
          (client) => client.userAgent = null;
    }

    deviceId ??= randomHexBytes(10);
    passportDeviceId ??= randomHexBytes(4);

    _dio.options.headers
      ..[HttpHeaders.userAgentHeader] =
          'iOS-14.7.1-7.4.203-iPhone10,2--$deviceId-${credential.userId}-$passportDeviceId-iPhone'
      ..['operate-common'] =
          '_region=CN&_language=ZH_CN&_deviceId=$deviceId&_appVersion=7.4.203&_platform=1&_platformVersion=14.7.1';

    _dio.options.headers[HttpHeaders.cookieHeader] =
        'PassportDeviceId=$passportDeviceId; '
        'locale=zh_CN; '
        'serviceToken=${credential.serviceToken}; '
        'userId=${credential.userId}; '
        'youpin_sessionid=${DateTime.now().millisecondsSinceEpoch}_00000000-0000-0000-0000-000000000000_223314141';
  }

  Future<MiIOServiceResponse<T>> request<T>(
    String url,
    Map<String, dynamic> data,
  ) async {
    final nonce = genNonce();
    final signedNonce = signNonce(
      nonce: nonce,
      ssecurity: credential.ssecurity,
    );

    final req = {
      'data': jsonEncode(data),
    };

    // Sign before entrypt.
    req['rc4_hash__'] = genSignature(
      signedNonce: signedNonce,
      url: url,
      params: req,
    );

    // Encrypt.
    final codec = MiRc4Codec(signedNonce);
    for (final entry in req.entries) {
      req[entry.key] = base64Encode(codec.convert(utf8.encode(entry.value)));
    }

    // Sign after entrypt.
    req['signature'] = genSignature(
      signedNonce: signedNonce,
      url: url,
      params: req,
    );

    req['_nonce'] = base64Encode(nonce);
    req['ssecurity'] = credential.ssecurity;

    logger.d('Post $url with payload:\n${jsonEncoder.convert(req)}');

    final resp = await _dio.post<String>(url, data: req);
    final respData = resp.data;
    if (respData == null) throw Exception('Failed to get response.');

    // Decrypt.
    var decrypted = MiRc4Codec(signedNonce).convert(base64Decode(respData));

    // Decompress if gzipped.
    if (decrypted[0] == 0x1F && decrypted[1] == 0x8B) {
      decrypted = gzip.decode(decrypted);
    }

    final respMap = jsonDecode(utf8.decode(decrypted)) as Map<String, dynamic>;
    logger.d('From $url got:\n${jsonEncoder.convert(respMap)}');

    return MiIOServiceResponse.fromJson(respMap);
  }

  /// Get devices registered to current account.
  Future<List<Map<String, dynamic>>> getDevices({
    bool getVirtualModel = true,
    bool getHuamiDevices = true,
    bool getSplitDevice = true,
  }) async {
    final resp = await request<Map<String, dynamic>>(
      '/app/v2/home/device_list_page',
      <String, dynamic>{
        'ssid': 'ChinaNet',
        'bssid': '44:F3:5E:B3:15:33',
        'getVirtualModel': getVirtualModel,
        'getHuamiDevices': getHuamiDevices ? 1 : 0,
        'get_split_device': getSplitDevice,
        'support_smart_home': true
      },
    );

    return (resp.result['list'] as List<dynamic>).cast();
  }

  /// Get a property.
  Future<T> getProperty<T>(String did, int siid, int piid) async {
    final resp = await getProperties([
      GetPropertyReq(did: did, siid: siid, piid: piid),
    ]);

    return resp.first.value as T;
  }

  /// Get a set of properties.
  Future<List<GetPropertyResp>> getProperties(
    List<GetPropertyReq> properties,
  ) async {
    final resp = await request<List<dynamic>>(
      '/app/miotspec/prop/get',
      <String, dynamic>{
        'params': properties,
      },
    );

    return resp.result.map((e) => GetPropertyResp.fromJson(e)).toList();
  }

  /// Set a property.
  Future<bool> setProperty<T>(String did, int siid, int piid, T value) async {
    final resp = await setProperties([
      SetPropertyReq<T>(did: did, siid: siid, piid: piid, value: value),
    ]);

    return resp.first.isOk;
  }

  /// Set a set of properties.
  Future<List<SetPropertyResp>> setProperties(
    List<SetPropertyReq> properties,
  ) async {
    final resp = await request<List<dynamic>>(
      '/app/miotspec/prop/set',
      <String, dynamic>{
        'params': properties,
      },
    );

    return resp.result.map((e) => SetPropertyResp.fromJson(e)).toList();
  }

  /// Trigger an action.
  Future<ActionResp> action(
    String did,
    int siid,
    int aiid, [
    List<dynamic> input = const <dynamic>[],
  ]) async {
    final resp = await request<Map<String, dynamic>>(
      '/app/miotspec/action',
      <String, dynamic>{
        'params': <String, dynamic>{
          'did': did,
          'siid': siid,
          'aiid': aiid,
          'in': input,
        }
      },
    );

    return ActionResp.fromJson(resp.result);
  }

  /// Get homes & rooms & associated devices.
  Future<Map<String, dynamic>> getHomes({
    bool fetchShare = true,
  }) async {
    final resp = await request<Map<String, dynamic>>(
      '/app/v2/homeroom/gethome',
      <String, dynamic>{
        "fg": true,
        "fetch_share": fetchShare,
        "fetch_share_dev": true,
        "limit": 300,
        "app_ver": 7
      },
    );

    return resp.result;
  }
}
