import 'dart:convert';
import 'dart:io';

import 'package:convert/convert.dart';
import 'package:cookie_jar/cookie_jar.dart';
import 'package:crypto/crypto.dart';
import 'package:dio/adapter.dart';
import 'package:dio/dio.dart';
import 'package:dio_cookie_manager/dio_cookie_manager.dart';

import 'model/credential.dart';
import 'utils.dart';

const _accountHost = 'account.xiaomi.com';
const _accountUrl = 'https://$_accountHost';

/// Login particular Mi service with username and password.
Future<MiServiceCredential> loginMiService(
  String sid,
  String username,
  String password, {
  String countryCode = '+86',
}) async {
  final deviceId = randomHexBytes(4);

  final cookieJar = DefaultCookieJar();
  cookieJar.domainCookies[_accountHost] = {
    '/': {
      'deviceId': SerializableCookie(Cookie('deviceId', deviceId)),
      'sdkVersion': SerializableCookie(Cookie('sdkVersion', '4.2.6')),
    }
  };

  final dio = Dio(BaseOptions(
    baseUrl: _accountUrl,
    headers: {
      HttpHeaders.acceptHeader: '*/*',
      HttpHeaders.acceptLanguageHeader: 'en-us',
      HttpHeaders.userAgentHeader:
          'APP/com.xiaomi.mihome APPV/7.4.203 iosPassportSDK/4.2.6 iOS/14.7.1 miHSTS',
    },
    contentType: 'application/x-www-form-urlencoded',
  ))
    ..interceptors.add(CookieManager(cookieJar));

  final httpClientAdapter = dio.httpClientAdapter;
  if (httpClientAdapter is DefaultHttpClientAdapter) {
    // Remove default UA.
    httpClientAdapter.onHttpClientCreate = (client) => client.userAgent = null;
  }

  final transformer = dio.transformer;
  if (transformer is DefaultTransformer) {
    // Remove prefix.
    transformer.jsonDecodeCallback =
        (str) => jsonDecode(str.replaceFirst('&&&START&&&', ''));
  }

  // Step 1
  var resp = await dio.get<Map<String, dynamic>>(
    '/pass/serviceLogin',
    queryParameters: <String, dynamic>{
      'sid': sid,
      '_json': true,
    },
  );

  logger.v('Mi service login step 1 got:\n${jsonEncoder.convert(resp.data)}');

  var data = resp.data;
  if (data == null || data['code'] != 70016) {
    throw Exception(data?['desc']);
  }

  final qs = data['qs'] as String;
  final sign = data['_sign'] as String;
  final callback = data['callback'] as String;

  // Step 2
  resp = await dio.post<Map<String, dynamic>>(
    '/pass/serviceLoginAuth2',
    data: <String, dynamic>{
      'user': username,
      '_json': 'true',
      'hash':
          hex.encode(md5.convert(utf8.encode(password)).bytes).toUpperCase(),
      'sid': sid,
      'callback': callback,
      '_sign': sign,
      'qs': qs,
      'cc': countryCode,
    },
  );

  logger.v('Mi service login step 2 got:\n${jsonEncoder.convert(resp.data)}');

  data = resp.data;
  if (data == null || data['code'] != 0) {
    throw Exception(data?['desc']);
  }

  final ssecurity = data['ssecurity'] as String;
  final location = data['location'] as String;

  // Step 3
  final resp3 = await dio.get<String>(location);
  logger.v('Mi service login step 3 got: $resp3');

  if (resp3.data != 'ok') {
    throw Exception(resp3.data);
  }

  final cookies = cookieJar.domainCookies[Uri.parse(callback).host]?['/'];
  final userId = cookies?['userId']?.cookie.value;
  final serviceToken = cookies?['serviceToken']?.cookie.value;

  if (userId == null || serviceToken == null) {
    throw Exception('Cannot read userId or serviceToken from cookie.');
  }

  return MiServiceCredential(
    sid: sid,
    userId: userId,
    ssecurity: ssecurity,
    serviceToken: serviceToken,
  );
}
